;;;
;;; Universal Monitor for uCOM-87
;;;   Copyright (C) 2021  Haruo Asano
;;;


	INCLUDE	"config.inc"

	INCLUDE	"../common.inc"

;;;
;;; ROM area
;;;

	ORG	0000H
	JMP	CSTART

	;; 
	;; CALT table
	;; 
	ORG	0080H
	;; 80-83  CALT 80H-
	DW	CSTART
	DW	WSTART0
	DW	CONOUT
	DW	STROUT
	;; 84-    CALT 88H-
	DW	CONIN
	DW	CONST

	CALT	88H

	ORG	0100H
CSTART:
	LXI	SP,STACK

	MVI	A,MM_V
	MOV	MM,A

	CALL	INIT

	LXI	H,RAM_B
	SHLD	DSADDR
	SHLD	SADDR
	SHLD	GADDR
	MVI	A,'I'
	MOV	HEXMOD,A

	;; Opening message
	LXI	H,OPNMSG
	CALL	STROUT
	JR	WSTART

WSTART0:
	POP	H		; Drop not needed return address
WSTART:
	LXI	H,PROMPT
	CALL	STROUT
	CALL	GETLIN
	LXI	H,INBUF
	CALL	SKIPSP
	CALL	UPPER
	ONI	A,0FFH
	JR	WSTART
	NEI	A,'D'
	JR	DUMP
	NEI	A,'G'
	JMP	GO
	NEI	A,'S'
	JMP	SETM

	NEI	A,'L'
	JMP	LOADH
	NEI	A,'P'
	JMP	SAVEH

ERR:
	LXI	H,ERRMSG
	CALL	STROUT
	JRE	WSTART

;;; Dump memory

DUMP:
	INX	H
	CALL	SKIPSP
	CALL	RDHEX		; 1st arg.
	MOV	A,C
	EQI	A,0
	JR	DP0
	;; No arg.
	CALL	SKIPSP
	LDAX	H
	EQI	A,0
	JR	ERR
	LHLD	DSADDR
	ADI	L,128
	ACI	H,0
	SHLD	DEADDR
	JRE	DPM
DP0:
	;; 1st arg. found
	SDED	DSADDR
	CALL	SKIPSP
	LDAX	H+
	NEI	A,','
	JR	DP1
	EQI	A,0
	JRE	ERR
	;; No 2nd arg.
	ADI	E,128
	ACI	D,0
	SDED	DEADDR
	JR	DPM
DP1:
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	NEI	C,0
	JRE	ERR
	LDAX	H
	EQI	A,0
	JRE	ERR
	INX	D
	SDED	DEADDR
DPM:
	;; DUMP main
	LHLD	DSADDR
	ANI	L,0F0H
	XRA	A,A
	MOV	DSTATE,A
DPM0:
	CALL	DPL
	CALL	CONST
	NEI	A,1
	JR	DPM1
	MOV	A,DSTATE
	EQI	A,2
	JR	DPM0
	LHLD	DEADDR
	SHLD	DSADDR
	JMP	WSTART
DPM1:
	CALL	CONIN
	SHLD	DSADDR
	JMP	WSTART

	;; Dump line
DPL:
	CALL	HEXOUT4
	PUSH	H
	LXI	H,DSEP0
	CALL	STROUT
	POP	H
	LXI	D,INBUF
	MVI	B,16-1
DPL0:
	CALL	DPB
	DCR	B
	JR	DPL0

	PUSH	H
	LXI	H,DSEP1
	CALL	STROUT

	;; Print ASCII area
	LXI	H,INBUF
	MVI	B,16-1
DPL1:
	LDAX	H+
	GTI	A,' '-1
	JR	DPL2
	LTI	A,7FH
	JR	DPL2
	CALL	CONOUT
	JR	DPL3
DPL2:
	MVI	A,'.'
	CALL	CONOUT
DPL3:
	DCR	B
	JR	DPL1
	POP	H
	JMP	CRLF

	;; Dump byte
DPB:
	MVI	A,' '
	CALL	CONOUT
	MOV	A,DSTATE
	EQI	A,0
	JR	DPB2
	;; Dump state 0
	MOV	A,DSADDR
	EQA	A,L
	JR	DPB0
	MOV	A,DSADDR+1
	NEA	A,H
	JR	DPB1
	;; Still 0 or 2
DPB0:
	MVI	A,' '
	CALL	CONOUT
	CALL	CONOUT
	STAX	D+
	INX	H
	RET
	;; Found start address
DPB1:
	MVI	A,1
	MOV	DSTATE,A
DPB2:
	MOV	A,DSTATE
	EQI	A,1
	JR	DPB0
	;; Dump state 1
	LDAX	H+
	STAX	D+
	CALL	HEXOUT2

	MOV	A,DEADDR
	EQA	A,L
	RET
	MOV	A,DEADDR+1
	EQA	A,H
	RET
	;; Found end address
	MVI	A,2
	MOV	DSTATE,A
	RET

;;; GO address

GO:
	INX	H
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	LDAX	H
	EQI	A,0
	JMP	ERR
	EQI	C,0
	SDED	GADDR
	LBCD	GADDR
	JB

;;; SET memory

SETM:
	INX	H
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	STAX	H
	EQI	A,0
	JMP	ERR
	PUSH	D
	POP	H
	EQI	C,0
	JR	SM0
	LHLD	SADDR
SM0:

SM1:
	CALL	HEXOUT4
	PUSH	H
	LXI	H,DSEP1
	CALL	STROUT
	POP	H
	LDAX	H
	PUSH	H
	CALL	HEXOUT2
	MVI	A,' '
	CALL	CONOUT
	CALL	GETLIN
	LXI	H,INBUF
	CALL	SKIPSP
	LDAX	H
	EQI	A,0
	JR	SM2
	;; Empty (Increment address)
	POP	H
	INX	H
	SHLD	SADDR
	JRE	SM1
SM2:
	EQI	A,'-'
	JR	SM3
	;; '-' (Decrement address)
	POP	H
	DCX	H
	SHLD	SADDR
	JRE	SM1
SM3:
	EQI	A,'.'
	JR	SM4
	;; '.' (Quit)
	POP	H
	SHLD	SADDR
	JMP	WSTART
SM4:
	CALL	RDHEX
	POP	H
	NEI	C,0
	JMP	ERR
	MOV	A,E
	STAX	H+
	SHLD	SADDR
	JRE	SM1

;;; LOAD HEX file

LOADH:
	INX	H
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	LDAX	H
	EQI	A,0
	JMP	ERR
LH0:
	CALL	CONIN
	CALL	UPPER
	NEI	A,'S'
	JMP	LHS0
LH1:
	NEI	A,':'
	JMP	LHI0
LH2:
	;; Skip to EOL
	NEI	A,CR
	JR	LH0
	NEI	A,LF
	JR	LH0
LH3:
	CALL	CONIN
	JR	LH2

LHI0:
	CALL	HEXIN
	MOV	C,A		; Checksum
	MOV	B,A		; Length

	CALL	HEXIN
	MOV	H,A		; Address H
	ADD	C,A		; Checksum

	CALL	HEXIN
	MOV	L,A		; Address L
	ADD	C,A		; Checksum

	;; Add offset
	MOV	A,E
	ADD	L,A
	MOV	A,D
	ADC	H,A

	CALL	HEXIN
	MOV	RECTYP,A
	ADD	C,A		; Checksum
	NEI	A,00H
	JR	LHI00
	NEI	A,01H
	JR	LHI00
	JRE	LH3		; Skip unsupported record type
LHI00:
	NEI	B,0
	JR	LHI3		; Length == 0
	DCR	B
LHI1:
	CALL	HEXIN
	ADD	C,A		; Checksum

	PUSH	V
	MOV	A,RECTYP
	EQI	A,0
	JR	LHI2
	POP	V
	STAX	H+
	JR	LHI20
LHI2:
	POP	V
LHI20:
	DCR	B
	JR	LHI1
LHI3:
	CALL	HEXIN
	ADD	A,C
	EQI	A,0
	JR	LHIE		; Checksum error
	MOV	A,RECTYP
	EQI	A,1
	JRE	LH3
	JMP	WSTART
LHIE:
	LXI	H,IHEMSG
	CALL	STROUT
	JMP	WSTART

LHS0:	
	CALL	CONIN
	NEI	A,'1'
	JR	LHS00
	NEI	A,'9'
	JR	LHS01
	JRE	LH3		; Unsupported record type
LHS00:
	MVI	A,0
LHS01:
	MVI	A,1
	MOV	RECTYP,A

	CALL	HEXIN
	MOV	B,A		; Length+3
	MOV	C,A		; Checksum

	CALL	HEXIN
	MOV	H,A		; Address H
	ADD	C,A		; Checksum

	CALL	HEXIN
	MOV	L,A		; Address L
	ADD	C,A		; Checksum

	;; Add offset
	MOV	A,E
	ADD	L,A
	MOV	A,D
	ADC	H,A

	GTI	B,3
	JR	LHS3
	SUI	B,3+1
LHS1:
	CALL	HEXIN
	ADD	C,A		; Checksum

	PUSH	V
	MOV	A,RECTYP
	EQI	A,0
	JR	LHS2
	POP	V
	STAX	H+
	JR	LHS20
LHS2:
	POP	V
LHS20:
	DCR	B
	JR	LHS1
LHS3:
	CALL	HEXIN
	ADD	A,C
	EQI	A,0FFH
	JR	LHSE		; Checksum error
	MOV	A,RECTYP
	EQI	A,1
	JRE	LH3
	JMP	WSTART
LHSE:
	LXI	H,SHEMSG
	CALL	STROUT
	JMP	WSTART

;;; SAVE HEX file

SAVEH:
	INX	H
	LDAX	H
	CALL	UPPER
	NEI	A,'I'
	JR	SH0
	EQI	A,'S'
	JR	SH1
SH0:
	MOV	HEXMOD,A
	INX	H
SH1:
	CALL	SKIPSP
	CALL	RDHEX
	NEI	C,0
	JMP	ERR
	SDED	PT0		; Start address
	CALL	SKIPSP
	LDAX	H+
	EQI	A,','
	JMP	ERR
	CALL	SKIPSP
	CALL	RDHEX
	NEI	C,0
	JMP	ERR
	CALL	SKIPSP
	LDAX	H
	EQI	A,0
	JMP	ERR
SH2:
	INX	D
	MOV	A,PT0
	SUB	E,A
	MOV	A,PT0+1
	SBB	D,A		; DE = Length
	LHLD	PT0		; HL = Start address
SH3:
	CALL	SHL
	EQI	D,0
	JR	SH3
	EQI	E,0
	JR	SH3

	MOV	A,HEXMOD
	EQI	A,'I'
	JR	SH4
	
	LXI	H,IHEXER
SH4:
	LXI	H,SRECER
	CALL	STROUT
	JMP	WSTART

SHL:
	MVI	B,16
	EQI	D,0
	JR	SHL0
	MOV	A,E
	LTA	A,B
	JR	SHL0
	MOV	B,A
SHL0:
	MOV	A,B
	SUB	E,A
	SBI	D,0

	MOV	A,HEXMOD
	EQI	A,'I'
	JRE	SHLS

	;; Intel HEX
	MVI	A,':'
	CALL	CONOUT

	MOV	A,B		; Length
	MOV	C,A		; Checksum
	CALL	HEXOUT2

	MOV	A,H		; Address H
	ADD	C,A		; Checksum
	CALL	HEXOUT2

	MOV	A,L		; Address L
	ADD	C,A		; Checksum
	CALL	HEXOUT2

	XRA	A,A
	CALL	HEXOUT2

	DCR	B
SHLI0:
	LDAX	H+
	ADD	C,A		; Checksum
	CALL	HEXOUT2

	DCR	B
	JR	SHLI0

	MOV	A,C
	NEGA
	CALL	HEXOUT2
	JMP	CRLF

SHLS:
	;; Motorola S record
	MVI	A,'S'
	CALL	CONOUT
	MVI	A,'1'
	CALL	CONOUT

	MOV	A,B
	ADI	A,2+1		; DataLength + Addr(2) + Sum(1)
	MOV	C,A		; Checksum
	CALL	HEXOUT2

	MOV	A,H		; Address H
	ADD	C,A		; Checksum
	CALL	HEXOUT2

	MOV	A,L		; Address L
	ADD	C,A		; Checksum
	CALL	HEXOUT2

	DCR	B
SHLS0:
	LDAX	H+
	ADD	C,A		; Checksum
	CALL	HEXOUT2

	DCR	B
	JR	SHLS0

	MOV	A,C
	XRI	A,0FFH
	CALL	HEXOUT2
	JMP	CRLF	

	;;
	;; Other support routines
	;;

STROUT:
	LDAX	H+
	NEI	A,0
	RET
	CALL	CONOUT
	JR	STROUT

HEXOUT4:
	MOV	A,H
	CALL	HEXOUT2
	MOV	A,L
HEXOUT2:
	PUSH	V
	SLR	A
	SLR	A
	SLR	A
	SLR	A
	CALL	HEXOUT1
	POP	V
HEXOUT1:
	ANI	A,0FH
	ADI	A,'0'
	GTI	A,'9'
	JMP	CONOUT
	ADI	A,'A'-'9'-1
	JMP	CONOUT

HEXIN:
	XRA	A,A
	CALL	HI0
	SLL	A
	SLL	A
	SLL	A
	SLL	A
HI0:
	PUSH	B
	MOV	C,A
	CALL	CONIN
	CALL	UPPER
	GTI	A,'0'-1
	JR	HIR
	GTI	A,'9'
	JR	HI1
	GTI	A,'A'-1
	JR	HIR
	LTI	A,'F'+1
	JR	HIR
	SUI	A,'A'-'9'-1
HI1:
	SUI	A,'0'
	ORA	A,C
HIR:
	POP	B
	RET

CRLF:
	MVI	A,CR
	CALL	CONOUT
	MVI	A,LF
	JMP	CONOUT

GETLIN:
	LXI	H,INBUF
	MVI	B,0
GL0:
	CALL	CONIN
	NEI	A,CR
	JRE	GLE
	NEI	A,LF
	JRE	GLE
	NEI	A,BS
	JR	GLB
	NEI	A,DEL
	JR	GLB
	GTI	A,' '-1
	JR	GL0
	LTI	A,80H
	JR	GL0
	MOV	C,A
	MOV	A,B
	LTI	A,BUFLEN-1
	JR	GL0		; Too long
	INR	B		; Never overflows
	MOV	A,C
	CALL	CONOUT
	STAX	H+
	JRE	GL0
GLB:
	MOV	A,B
	NEI	A,0
	JRE	GL0
	DCR	B		; Never overflows
	DCX	H
	MVI	A,BS
	CALL	CONOUT
	MVI	A,' '
	CALL	CONOUT
	MVI	A,BS
	CALL	CONOUT
	JRE	GL0
GLE:
	CALL	CRLF
	XRA	A,A
	STAX	H
	RET

SKIPSP:
	LDAX	H
	EQI	A,' '
	RET
	INX	H
	JR	SKIPSP

UPPER:
	GTI	A,'a'-1
	RET
	LTI	A,'z'+1
	RET
	ADI	A,'A'-'a'
	RET

RDHEX:
	MVI	E,0
	LXI	B,0
RH0:
	LDAX	H+
	CALL	UPPER
	GTI	A,'0'-1
	JRE	RHE
	GTI	A,'9'
	JR	RH1
	GTI	A,'A'-1
	JRE	RHE
	LTI	A,'F'+1
	JRE	RHE
	SUI	A,'A'-'9'-1
RH1:
	SUI	A,'0'
	RLL	A
	RLL	A
	RLL	A
	RLL	A
	RLL	A
	RLL	C
	RLL	B
	RLL	A
	RLL	C
	RLL	B
	RLL	A
	RLL	C
	RLL	B
	RLL	A
	RLL	C
	RLL	B
	ADI	E,1
	JRE	RH0
RHE:
	DCX	H
	PUSH	B
	MOV	A,E
	MOV	C,A
	POP	D
	RET

OPNMSG:
	DB	CR,LF,"Universal Monitor uCOM-87",CR,LF,00H

PROMPT:
	DB	"] ",00H

IHEMSG:
	DB	"Error ihex",CR,LF,00H

SHEMSG:
	DB	"Error srec",CR,LF,00H

ERRMSG:
	DB	"Error",CR,LF,00H

DSEP0:
	DB	" :",00H
DSEP1:
	DB	" : ",00H
IHEXER:
        DB	":00000001FF",CR,LF,00H
SRECER:
        DB	"S9030000FC",CR,LF,00H

	IF USE_DEV_7810
	INCLUDE	"dev/dev_7810.asm"
	ENDIF

	IF USE_DEV_EMILY
	INCLUDE	"dev/dev_emily.asm"
	ENDIF

;;;
;;; RAM area
;;;
	
	;;
	;; Work area
	;;

	ORG	WORK_B

INBUF:	DS	BUFLEN		; Line input buffer
DSADDR:	DS	2		; Dump start address
DEADDR:	DS	2		; Dump end address
DSTATE:	DS	1		; Dump state
GADDR:	DS	2		; Go address
SADDR:	DS	2		; Set address
HEXMOD:	DS	1		; HEX file mode
RECTYP:	DS	1		; Record Type

PT0:	DS	2		; Temporary pointer

	IFDEF MEMREQ
DEVMEM:	DS	MEMREQ
	ENDIF

	END
